#include <antiglut.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

static char *
event_type_name(ui_event_p ev)
{
	switch (ev->type)
	{
		case ui_key_event:
			return "ui_key_event";
		case ui_pointer_motion_event:
			return "ui_pointer_motion_event";
		case ui_window_focus_in_event:
			return "ui_window_focus_in_event";
		case ui_window_focus_out_event:
			return "ui_window_focus_out_event";
		case ui_window_expose_event:
			return "ui_window_expose_event";
		case ui_window_map_event:
			return "ui_window_map_event";
		case ui_window_unmap_event:
			return "ui_window_unmap_event";
		case ui_window_move_event:
			return "ui_window_move_event";
		case ui_window_size_event:
			return "ui_window_size_event";
		case ui_wm_event:
			return "ui_wm_event";
		default:
			break;
	}
	return "unknown";
}

static char *
event_scan_name(ui_event_t *ev)
{
	switch(ev->key.scan)
	{
		case sc_escape: return "sc_escape";
		case sc_f1: return "sc_f1";
		case sc_f2: return "sc_f2";
		case sc_f3: return "sc_f3";
		case sc_f4: return "sc_f4";
		case sc_f5: return "sc_f5";
		case sc_f6: return "sc_f6";
		case sc_f7: return "sc_f7";
		case sc_f8: return "sc_f8";
		case sc_f9: return "sc_f9";
		case sc_f10: return "sc_f10";
		case sc_f11: return "sc_f11";
		case sc_f12: return "sc_f12";
		case sc_print: return "sc_print";
		case sc_scroll_lock: return "sc_scroll_lock";
		case sc_pause: return "sc_pause";
	
		case sc_grave: return "sc_grave";
		case sc_1: return "sc_1";
		case sc_2: return "sc_2";
		case sc_3: return "sc_3";
		case sc_4: return "sc_4";
		case sc_5: return "sc_5";
		case sc_6: return "sc_6";
		case sc_7: return "sc_7";
		case sc_8: return "sc_8";
		case sc_9: return "sc_9";
		case sc_0: return "sc_0";
		case sc_minus: return "sc_minus";
		case sc_equal: return "sc_equal";
		case sc_backslash: return "sc_backslash";
		case sc_backspace: return "sc_backspace";
	
		case sc_tab: return "sc_tab";
		case sc_q: return "sc_q";
		case sc_w: return "sc_w";
		case sc_e: return "sc_e";
		case sc_r: return "sc_r";
		case sc_t: return "sc_t";
		case sc_y: return "sc_y";
		case sc_u: return "sc_u";
		case sc_i: return "sc_i";
		case sc_o: return "sc_o";
		case sc_p: return "sc_p";
		case sc_lbracket: return "sc_lbracket";
		case sc_rbracket: return "sc_rbracket";
		case sc_enter: return "sc_enter";
	
		case sc_caps_lock: return "sc_caps_lock";
		case sc_a: return "sc_a";
		case sc_s: return "sc_s";
		case sc_d: return "sc_d";
		case sc_f: return "sc_f";
		case sc_g: return "sc_g";
		case sc_h: return "sc_h";
		case sc_j: return "sc_j";
		case sc_k: return "sc_k";
		case sc_l: return "sc_l";
		case sc_semicolon: return "sc_semicolon";
		case sc_apostrophe: return "sc_apostrophe";
	
		case sc_lshift: return "sc_lshift";
		case sc_z: return "sc_z";
		case sc_x: return "sc_x";
		case sc_c: return "sc_c";
		case sc_v: return "sc_v";
		case sc_b: return "sc_b";
		case sc_n: return "sc_n";
		case sc_m: return "sc_m";
		case sc_comma: return "sc_comma";
		case sc_period: return "sc_period";
		case sc_slash: return "sc_slash";
		case sc_rshift: return "sc_rshift";
	
		case sc_lcontrol: return "sc_lcontrol";
		case sc_lsuper: return "sc_lsuper";
		case sc_lalt: return "sc_lalt";
		case sc_space: return "sc_space";
		case sc_ralt: return "sc_ralt";
		case sc_rsuper: return "sc_rsuper";
		case sc_menu: return "sc_menu";
		case sc_rcontrol: return "sc_rcontrol";
	
		case sc_insert: return "sc_insert";
		case sc_home: return "sc_home";
		case sc_prior: return "sc_prior";
		case sc_delete: return "sc_delete";
		case sc_end: return "sc_end";
		case sc_next: return "sc_next";
	
		case sc_up: return "sc_up";
		case sc_down: return "sc_down";
		case sc_left: return "sc_left";
		case sc_right: return "sc_right";
	
		case sc_num_lock: return "sc_num_lock";
		case sc_kp_divide: return "sc_kp_divide";
		case sc_kp_multiply: return "sc_kp_multiply";
		case sc_kp_subtract: return "sc_kp_subtract";
		case sc_kp_home: return "sc_kp_home";
		case sc_kp_up: return "sc_kp_up";
		case sc_kp_prior: return "sc_kp_prior";
		case sc_kp_add: return "sc_kp_add";
		case sc_kp_left: return "sc_kp_left";
		case sc_kp_begin: return "sc_kp_begin";
		case sc_kp_right: return "sc_kp_right";
		case sc_kp_end: return "sc_kp_end";
		case sc_kp_down: return "sc_kp_down";
		case sc_kp_next: return "sc_kp_next";
		case sc_kp_enter: return "sc_kp_enter";
		case sc_kp_insert: return "sc_kp_insert";
		case sc_kp_delete: return "sc_kp_delete";
		
		case sc_lbutton: return "sc_lbutton";
		case sc_mbutton: return "sc_mbutton";
		case sc_rbutton: return "sc_rbutton";
	
		case sc_wheelup: return "sc_wheelup";
		case sc_wheeldown: return "sc_wheeldown";
		default: break;
	}
	return "unknown";
}

static char *
get_buttons(ui_event_p ev)
{
	static char btns[512];
	btns[0] = 0;
	if(ev->button & UI_LBUTTON_MASK)
		strcat(btns,"UI_LBUTTON_MASK");
	if(ev->button & UI_MBUTTON_MASK)
	{
		if(btns[0])
			strcat(btns,"|");
		strcat(btns,"UI_MBUTTON_MASK");
	}
	if(ev->button & UI_RBUTTON_MASK)
	{
		if(btns[0])
			strcat(btns,"|");
		strcat(btns,"UI_RBUTTON_MASK");
	}
	return btns;
}

static char *
get_mods(ui_event_p ev)
{
	static char mods[512];
	mods[0] = 0;
	if(ev->state & UI_NUM_LOCK_MASK)
		strcat(mods,"UI_NUM_LOCK_MASK");
	if(ev->state & UI_CAPS_LOCK_MASK)
	{
		if(mods[0])
			strcat(mods,"|");
		strcat(mods,"UI_CAPS_LOCK_MASK");
	}
	if(ev->state & UI_SCROLL_LOCK_MASK)
	{
		if(mods[0])
			strcat(mods,"|");
		strcat(mods,"UI_SCROLL_LOCK_MASK");
	}
	if(ev->state & UI_SHIFT_MASK)
	{
		if(mods[0])
			strcat(mods,"|");
		strcat(mods,"UI_SHIFT_MASK");
	}
	if(ev->state & UI_CONTROL_MASK)
	{
		if(mods[0])
			strcat(mods,"|");
		strcat(mods,"UI_CONTROL_MASK");
	}
	if(ev->state & UI_ALT_MASK)
	{
		if(mods[0])
			strcat(mods,"|");
		strcat(mods,"UI_ALT_MASK");
	}
	if(ev->state & UI_SUPER_MASK)
	{
		if(mods[0])
			strcat(mods,"|");
		strcat(mods,"UI_SUPER_MASK");
	}
	return mods;
}

static char *
show_config_event(ui_event_p ev)
{
	static char buf[4096];
	int tail = 0;
	tail = sprintf(buf,"Window config: Focus=");
	if(ev->configure.focused)
		tail += sprintf(buf+tail,"YES;");
	else
		tail += sprintf(buf+tail,"NO;");
	tail += sprintf(buf+tail,"Visibility=");
	if(ev->configure.visible)
		tail += sprintf(buf+tail,"YES;");
	else
		tail += sprintf(buf+tail,"NO;");
	tail += sprintf(buf+tail,"Damaged=[%u,%u,%u,%u];",
		ev->configure.exposed_left, ev->configure.exposed_top,
		ev->configure.exposed_right, ev->configure.exposed_bottom);
	tail += sprintf(buf+tail,"PosUL=[%u,%u];",
		ev->configure.x,ev->configure.y);
	tail += sprintf(buf+tail,"Size=[%u,%u];",
		ev->configure.width,ev->configure.height);
	sprintf(buf+tail,"WM code=[%u]\n",ev->configure.wm_event);
	return buf;
}

static void
display_event(ui_event_p ev)
{
	unsigned i;
	printf("Type: '%s' (%u)\n",event_type_name(ev),ev->type);
	printf("Time: %g\n",ev->time);
	printf("Client X=%d,Y=%d\n",ev->client_x,ev->client_y);
	printf("Screen X=%d,Y=%d\n",ev->screen_x,ev->screen_y);
	printf("Client DX=%d,DY=%d\n",ev->client_dx,ev->client_dy);
	printf("Mouse buttons: '%s'\n",get_buttons(ev));
	printf("Modifiers: '%s'\n",get_mods(ev));
	switch(ev->type)
	{
		case ui_key_event:
			printf("%s: Scan=%s,ASCII=%u,%02X (%c),UNI(ME16)=%04X;UTF8=",
				ev->key.press ? "Pressed" : "Released",
				event_scan_name(ev),
				ev->key.acode,ev->key.acode,(ev->key.acode >= 32) ? ev->key.acode : '?',
				ev->key.ucode);
			for(i=0;i < ev->key.utf8_len;i++)
				printf("(%02X)",(unsigned)(ev->key.utf8[i]));
			puts(".");
			break;
		case ui_pointer_motion_event:
			break;
		default:
			puts(show_config_event(ev));
	}
	puts("================================================================================");
}

#define	MINDELTA	33	/* 30 FPS max */

static void
draw(void)
{
	static unsigned char count = 0;
	unsigned newtick;
	static int flip = 0;
	static unsigned tick = 0;
	GLfloat cl = count / 255.0f;
	GLfloat phase = (GLfloat)sin(2 * 3.1416f * count / 255.0f);
	GLfloat sc = 0.9f + 0.05f * phase;
	switch(flip)
	{
		case 0: glClearColor(0,cl,cl,0.5f); break;
		case 1: glClearColor(cl,0,cl,0.5f); break;
		case 2: glClearColor(cl,cl,0,0.5f); break;
	}
	if(!count++)
		flip = (flip + 1) % 3;
	glClear(GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glTranslatef(0.5f,0.5f,0);
	glRotatef(20 * phase,0,0,1);
	glTranslatef(-1,-1,0);
	glScalef(2,2,2);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glScalef(sc,sc,sc);
	glColor4f(0.7f,0.5f,0.8f,0.5f);
	glBegin(GL_QUADS);
		glTexCoord2f(1,1);
		glVertex2f(1,1);
		glTexCoord2f(0,1);
		glVertex2f(-1,1);
		glTexCoord2f(0,0);
		glVertex2f(-1,-1);
		glTexCoord2f(1,0);
		glVertex2f(1,-1);
	glEnd();
	agSwap();

	newtick = agTickCount();
	/* check for overflow */
	if(newtick >= tick && (newtick - tick) < MINDELTA)
		agSleep(MINDELTA - (newtick - tick));
	tick = newtick;
}

static void
upload()
{
	static GLfloat tex[] = {
		1,.5f,1,.6f,
		.3f,1,.2f,1,
		1,.2f,1,.3f,
		.6f,1,.5f,1
	};
	glEnable(GL_TEXTURE_2D);
	glTexImage2D(GL_TEXTURE_2D,0,1,4,4,0,GL_LUMINANCE,GL_FLOAT,tex);
	glTexEnvf(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
}

int main()
{
	GLfloat aspect;
	ui_event_t ev;
	agWindowedConfig("AntiGlut: event demo",512,512,40,20);
	if(!agInit(AG_REQUIRED_1_1,GL_FALSE,GL_FALSE,GL_FALSE))
		return -1;
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-1,1,-1,1,-1,1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	upload();
	uiAutoRepeat(GL_FALSE);
	for(;;)
	{
		while(uiPending())
		{
			uiNextEvent(&ev);
			display_event(&ev);
			switch(ev.type)
			{
				case ui_wm_event:
					if(ev.configure.wm_event == ui_wm_close)
						goto done;
					break;
				case ui_window_size_event:
					glViewport(0,0,ev.configure.width,ev.configure.height);
					glMatrixMode(GL_PROJECTION);
					glLoadIdentity();
					aspect = (GLfloat)ev.configure.width / ev.configure.height;
					glOrtho(-aspect,aspect,-1,1,-1,1);
					glMatrixMode(GL_MODELVIEW);
					break;
				default:
					break;
			}
		}
		draw();
	}
done:;
	agShutdown();
	return 0;
}
